import datetime
import logging
from multiprocessing import Process, Queue
from typing import Text

from apscheduler.job import Job
from apscheduler.schedulers.background import BackgroundScheduler
from pytz import UnknownTimeZoneError, utc
from rasax.community import utils, config
from rasax.community.services.analytics_service import AnalyticsService

logger = logging.getLogger(__name__)

_job_queue = Queue()


def _schedule_background_jobs(scheduler: BackgroundScheduler) -> None:
    from rasax.community.services.git_service import GitService, GIT_BACKGROUND_JOB_ID

    # Schedule Git synchronization
    scheduler.add_job(
        GitService.run_background_synchronization,
        "cron",
        id=GIT_BACKGROUND_JOB_ID,
        next_run_time=datetime.datetime.now(),
        replace_existing=True,
        minute="*",
    )

    # Schedule analytics caching
    if utils.is_enterprise_installed():
        scheduler.add_job(
            AnalyticsService.run_analytics_caching,
            "cron",
            replace_existing=True,
            **config.analytics_update_kwargs,
        )


def _run_scheduler():
    try:
        logging.getLogger("apscheduler.scheduler").setLevel(logging.WARNING)
        scheduler = BackgroundScheduler()
        scheduler.start()
    except UnknownTimeZoneError:
        logger.warning(
            "apscheduler could not find a timezone and is "
            "defaulting to utc. This is probably because "
            "your system timezone is not set. "
            'Set it with e.g. echo "Europe/Berlin" > '
            "/etc/timezone"
        )
        scheduler = BackgroundScheduler(timezone=utc)
        scheduler.start()

    _schedule_background_jobs(scheduler)

    # Check regularly if a job should be executed right away
    while True:
        job_id = _job_queue.get()
        existing_job = scheduler.get_job(job_id)  # type: Job
        if existing_job:
            logger.debug(f"Running job with id '{job_id}'' immediately.")
            scheduler.modify_job(job_id, next_run_time=datetime.datetime.now())
        else:
            logger.warning(f"Did not find a scheduled job with id '{job_id}'.")


def run_job_immediately(job_id: Text) -> None:
    """Trigger a scheduled background job to run immediately.
    
    Args:
        job_id: Id of the job which should be triggered.
    """
    _job_queue.put(job_id)


def start_background_scheduler() -> Process:
    """Start a background scheduler which runs periodic tasks."""

    # Start scheduler in a separate process so that we can create a process and thread
    # safe interface by using a `Queue` to communicate with it.
    process = Process(target=_run_scheduler)
    process.start()

    return process
