import logging
import os
from pathlib import Path
from typing import Text, Dict, Union, Optional, List

import rasax.community.initialise
import rasax.community.services.data_service
from git import Repo, GitCommandError
from rasax.community import config, utils
from sqlalchemy.orm import Session
from rasa.utils import io as io_utils

from rasax.community.constants import DEFAULT_GIT_REPOSITORY_DIRECTORY
from rasax.community.database.admin import GitRepository
from rasax.community.database.service import DbService

logger = logging.getLogger(__name__)

DEFAULT_TARGET_BRANCH = "master"

DEFAULT_REMOTE_NAME: Text = "origin"
GIT_BACKGROUND_JOB_ID = "git-synchronization"
SSH_FILES_DIRECTORY = "ssh_files"


class GitService(DbService):
    def __init__(
        self,
        session: Session,
        directory_for_git_clones: Union[Text, Path] = DEFAULT_GIT_REPOSITORY_DIRECTORY,
        project_id: Text = config.project_name,
        repository_id: Optional[int] = None,
    ) -> None:
        """Creates a `GitService` instance for a given repository.

        Args:
            session: The database session.
            directory_for_git_clones: Parent directory which should be used to store
                the cloned repositories.
            project_id: The ID of the project.
            repository_id: The ID of the repository which this service should be
                used with.
        """

        super().__init__(session)
        self._directory_for_git_clones = Path(directory_for_git_clones)
        io_utils.create_directory(directory_for_git_clones)

        self._project: Text = project_id

        self._repository: Optional[Repo] = None
        self._repository_id: Optional[int] = None
        if repository_id:
            self.set_repository(repository_id)

    def _target_branch(self) -> Text:
        return self._get_repository(self._repository_id).target_branch

    def get_current_branch(self) -> Text:
        """Get the currently active Git branch.

        Returns:
            Name of the currently active branch.
        """

        return self._repository.active_branch.name

    def create_branch(self, branch_name: Text) -> None:
        """Creates a new branch (does not check it out).

        Args:
            branch_name: Name of the branch.
        """

        self._repository.create_head(branch_name)

    def checkout_branch(
        self, branch_name: Text, force: bool = False, inject_changes: bool = True
    ) -> None:
        """Checks out a branch (`git checkout`).

        Args:
            branch_name: Name of the branch which should be checked out.
            force: If `True` discard any local changes before switching branch.
            inject_changes: If `True` the latest data is injected into the database.
        """

        if self.is_dirty() and force:
            self._repository.git.reset("--hard")

        matching_branch = next(
            (
                branch
                for branch in self._repository.branches
                if branch.name == branch_name
            ),
            None,
        )
        if not matching_branch:
            raise ValueError(f"Branch '{matching_branch}' does not exist.")

        matching_branch.checkout()

        if inject_changes:
            self.trigger_immediate_project_synchronization()

    def is_dirty(self) -> bool:
        """Checks if there are uncommitted changes.

        Returns:
            `True` if there are uncommitted changes, otherwise `False`.
        """

        return self._repository.is_dirty(untracked_files=True)

    def save_repository(
        self, repository_information: Dict[Text, Union[Text, int]]
    ) -> Dict[Text, Text]:
        """Saves the credentials for a repository.

        Args:
            repository_information: The information for the repository as `Dict`.

        Returns:
            The saved repository information including database ID.
        """

        repository_credentials = GitRepository(
            name=repository_information.get("name"),
            repository_url=repository_information["repository_url"],
            ssh_key=repository_information.get("ssh_key"),
            git_service=repository_information.get("git_service"),
            git_service_access_token=repository_information.get(
                "git_service_access_token"
            ),
            target_branch=repository_information.get(
                "target_branch", DEFAULT_TARGET_BRANCH
            ),
            project_id=self._project,
        )

        self.add(repository_credentials)

        self.flush()  # to assign database id

        if repository_credentials.ssh_key:
            self._add_ssh_key(repository_credentials.ssh_key, repository_credentials.id)

        io_utils.create_directory(self.repository_path(repository_credentials.id))
        self.set_repository(repository_credentials.id)

        return repository_credentials.as_dict()

    def _add_ssh_key(self, ssh_key: Text, repository_id: int) -> None:
        base_path = self._directory_for_git_clones / SSH_FILES_DIRECTORY
        io_utils.create_directory(base_path)
        path_to_key = base_path / f"{repository_id}.key"
        utils.write_text_to_file(path_to_key, f"{ssh_key}\n")

        # We have to be able to read and write this
        path_to_key.chmod(0o600)

        command = f"""
#!/bin/sh
ID_RSA={path_to_key}
exec /usr/bin/ssh -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no -i $ID_RSA "$@"        
        """

        path_to_executable = self._ssh_executable_path(repository_id)
        utils.write_text_to_file(path_to_executable, command)

        # We have to be able read, write and execute the file
        path_to_executable.chmod(0o700)

    def _ssh_executable_path(self, repository_id: int) -> Path:
        return (
            self._directory_for_git_clones / SSH_FILES_DIRECTORY / f"{repository_id}.sh"
        )

    def get_repositories(self) -> List[Dict[Text, Union[Text, int]]]:
        """Retrieves the stored repository information from the database.

        Returns:
            Stored git credentials (not including ssh keys and access tokens).
        """

        repositories = (
            self.query(GitRepository)
            .filter(GitRepository.project_id == self._project)
            .order_by(GitRepository.id)
            .all()
        )
        serialized_repositories = [credential.as_dict() for credential in repositories]

        return serialized_repositories

    def get_most_recent_repository(self) -> Optional[GitRepository]:
        """Returns the credentials for the current working directory.

        Returns:
            `None` if no repository information has been stored, otherwise a `Dict` of the
            provided repository details (not including ssh keys and access tokens)..
        """

        return (
            self.session.query(GitRepository).order_by(GitRepository.id.desc()).first()
        )

    def get_repository(self) -> Dict[Text, Union[Text, int]]:
        """Retrieve the repository information for the current repository."""

        return self._get_repository(self._repository_id).as_dict()

    def _get_repository(self, repository_id: int) -> GitRepository:
        result = (
            self.query(GitRepository).filter(GitRepository.id == repository_id).first()
        )
        if not result:
            raise ValueError(f"Repository with ID '{repository_id}' does not exist.")
        else:
            return result

    def update_repository(
        self, repository_information: Dict[Text, Union[Text, int]]
    ) -> None:
        """Updates the current repository with new credentials.

        Args:
            repository_information: The values which overwrite the currently
                stored information.
        """

        old = self._get_repository(self._repository_id)
        keys = [
            "name",
            "repository_url",
            "ssh_key",
            "git_service",
            "git_service_access_token",
            "target_branch",
        ]
        for key in keys:
            setattr(old, key, repository_information[key])

    def delete_repository(self) -> None:
        """Deletes the current repository from the database and the file system."""

        import shutil

        result = self._get_repository(self._repository_id)
        self.delete(result)

        # Delete cloned directory
        shutil.rmtree(self.repository_path())

    def clone(self) -> None:
        """Clone the current repository."""

        repository_info = self._get_repository(self._repository_id)
        target_path = self.repository_path(repository_info.id)

        if self._repository or self._is_git_directory(target_path):
            raise ValueError(f"'{target_path}' is already a git directory.")

        repository_url = repository_info.repository_url
        logger.info(f"Cloning git repository from URL '{repository_url}'.")

        # Configure authentication
        # TODO: Add option to use token for cloning
        # (https://help.github.com/en/github/authenticating-to-github/creating-a-personal-access-token-for-the-command-line)
        environment = {}
        if repository_info.ssh_key:
            self._add_ssh_key(repository_info.ssh_key, repository_info.id)
            environment["GIT_SSH"] = str(self._ssh_executable_path(repository_info.id))

        self._repository_id = repository_info.id
        # Actually clone it
        try:
            self._repository = Repo.clone_from(
                repository_url, target_path, env=environment
            )
        except GitCommandError as e:
            logger.error(
                f"Failed to clone repository from URL '{repository_url}'.\n" f"{e}"
            )

        logger.debug(f"Finished cloning git repository '{repository_url}'.")

    def repository_path(self, repository_id: Optional[int] = None) -> Path:
        """Returns the path to the current repository on disk.

        Args:
            repository_id: If a repository id is given, it returns the path to this
                repository instead of the current one.

        Returns:
            Path to repository on disk.
        """

        if not repository_id and self._repository_id:
            repository_id = self._repository_id

        if repository_id:
            return self._directory_for_git_clones / str(repository_id)
        else:
            raise ValueError("No path to Git repository found.")

    def set_repository(self, repository_id: int) -> None:
        """Changes the current repository of the service to another one.

        Args:
            repository_id: ID of the repository which should be used.
        """

        directory = self._directory_for_git_clones / str(repository_id)
        self._repository_id = repository_id

        if self._is_git_directory(directory):
            self._repository = Repo(directory)
            repository = self._get_repository(repository_id)

            if repository.ssh_key:
                self._repository.git.update_environment(
                    GIT_SSH=str(self._ssh_executable_path(self._repository_id))
                )
        else:
            self._repository = None

    @staticmethod
    def _is_git_directory(directory: Union[Text, Path]) -> bool:
        return os.path.exists(directory) and ".git" in os.listdir(directory)

    def commit_to(
        self, branch_name: Text, message: Optional[Text]
    ) -> Dict[Text, Union[Text, bool]]:
        """Commit local changes and push them to remote.

        Args:
            branch_name: Branch which the changes should be committed to.
            message: Commit message.

        Returns:
            Information about the commit.
        """

        if branch_name == self._target_branch():
            _ = self.commit_to_target_branch(message)
            return {"remote_branch": branch_name, "committed_to_target_branch": True}
        else:
            _ = self.commit_to_new_branch(branch_name, message)
            return {"remote_branch": branch_name, "committed_to_target_branch": False}

    def commit_to_target_branch(self, message: Optional[Text]) -> Text:
        """Commits changes to the target branch.

        The target branch is e.g. `master` for 'production environment'.

        Args:
            message: The commit message. If `None` a message is generated.

        Returns:
            Name of the target branch.
        """

        self._commit_changes(message)
        return self._target_branch()

    def _commit_changes(self, message: Optional[Text]) -> None:
        self._repository.git.add(A=True)
        self._repository.index.commit(message)

    def commit_to_new_branch(self, branch_name: Text, message: Optional[Text]) -> Text:
        """Commits changes to a new branch.

        Args:
            message: The commit message. If `None` a message is generated.
            branch_name: Branch which should be used.

        Returns:
            Used branch name (the name will be generated).
        """

        self.create_branch(branch_name)
        self.checkout_branch(branch_name, inject_changes=False)
        self._commit_changes(message)
        self.checkout_branch(self._target_branch())

        return branch_name

    def merge_latest_remote_branch(self) -> None:
        """Merges the current branch with its remote branch."""

        self._repository.git.fetch()

        current_branch = self.get_current_branch()
        if not self.is_dirty() and self.is_remote_branch_ahead(current_branch):
            logger.debug(
                f"Merging remote branch '{current_branch}' into current branch '{current_branch}'."
            )
            self._repository.git.merge(self._remote_branch_name(current_branch))

    def _remote_branch_name(self, branch: Text) -> Text:
        return f"{DEFAULT_REMOTE_NAME}/{branch}"

    def is_remote_branch_ahead(self, other_branch: Optional[Text] = None) -> bool:
        """Checks if `other_branch` has new commits.

        Args:
            other_branch: Remote branch which should be used for comparison. If `None`,
                the remote target branch is used.

        Returns:
            `True` if the remote branch is ahead, `False` if it's not ahead.
        """

        if not other_branch:
            other_branch = self._target_branch()

        current_branch = self.get_current_branch()
        remote_branch = self._remote_branch_name(other_branch)
        commits_behind = self._repository.iter_commits(
            f"{current_branch}..{remote_branch}"
        )
        number_of_commits_behind = sum(1 for _ in commits_behind)

        logger.debug(
            f"Branch '{current_branch}' is {number_of_commits_behind} commits behind '{other_branch}'."
        )
        return number_of_commits_behind > 0

    def push(self, branch_name: Text) -> None:
        """Push committed changes.

        Args:
            branch_name: Branch which should be pushed.
        """

        self._repository.remote().push(branch_name)

    async def synchronize_project(self) -> None:
        """Synchronizes the Git repository with the database."""

        repository_info = self.get_most_recent_repository()
        if not repository_info:
            logger.debug(
                "Skip synchronizing with Git since no credentials where given."
            )
            return

        # Git directory already exists - let's try to fetch the changes
        self.set_repository(repository_info.id)
        repository_path = self.repository_path(repository_info.id)

        if not self._is_git_directory(repository_path):
            self.clone()
            await self._inject_data()
            return

        if self.is_dirty():
            logger.debug(
                "Skip synchronizing with Git since the working directory "
                "contains changes. Please commit and push them in order to "
                "pull the latest changes."
            )
            return
        elif not self.is_remote_branch_ahead():
            logger.debug(
                "Remote branch does not contain new changes. No new data "
                "needs to be injected."
            )
            return
        else:
            self.merge_latest_remote_branch()
            await self._inject_data()

    async def _inject_data(self) -> None:
        from rasax.community import local
        from rasa.constants import DEFAULT_DATA_PATH

        data_path = self.repository_path() / DEFAULT_DATA_PATH
        logger.debug("Injecting latest changes from git.")

        await rasax.community.initialise.inject_files_from_disk(
            str(self.repository_path()), str(data_path), self.session
        )

    def get_repository_status(self) -> Dict[Text, Union[Text, bool]]:
        """Returns the current repository status."""

        is_remote_ahead = self.is_remote_branch_ahead()

        return {
            "is_committing_to_target_branch_allowed": not is_remote_ahead,
            "is_remote_ahead": is_remote_ahead,
            "are_there_local_changes": self.is_dirty(),
            "current_branch": self.get_current_branch(),
            "time_of_last_pull": self._get_latest_pull_time(),
        }

    def _get_latest_pull_time(self) -> float:
        fetch_head_path = Path(self._repository.git_dir) / "FETCH_HEAD"
        if fetch_head_path.exists():
            return fetch_head_path.stat().st_mtime
        else:
            logging.debug(
                f"Failed to fetch the latest repository pull time for the repository "
                f"with ID {self._repository_id}. Using the timestamp of the latest "
                f"commit instead."
            )
            return self._repository.commit().committed_datetime.timestamp()

    @staticmethod
    def run_background_synchronization() -> None:
        from rasax.community.database import utils as db_utils
        import asyncio

        logger.debug("Running scheduled Git synchronization.")

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        with db_utils.session_scope() as session:
            git_service = GitService(session, DEFAULT_GIT_REPOSITORY_DIRECTORY)
            loop.run_until_complete(git_service.synchronize_project())

    @staticmethod
    def trigger_immediate_project_synchronization() -> None:
        """Triggers an immediate synchronization of the Git repository and the database.
        """

        from rasax.community import scheduler

        scheduler.run_job_immediately(GIT_BACKGROUND_JOB_ID)
